package cz.fidentis.analyst.visitors.mesh;

import cz.fidentis.analyst.mesh.MeshVisitor;
import cz.fidentis.analyst.mesh.core.MeshFacet;
import cz.fidentis.analyst.mesh.core.MeshPoint;
import cz.fidentis.analyst.mesh.core.MeshTriangle;
import cz.fidentis.analyst.mesh.core.TriangleFan;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.vecmath.Vector3d;

/**
 * Abstract class for algorithms calculating curvatures of mesh vertices.
 * @see https://computergraphics.stackexchange.com/questions/1718/what-is-the-simplest-way-to-compute-principal-curvature-for-a-mesh-triangle
 * @see http://rodolphe-vaillant.fr/?e=20
 * <p>
 * <b>All curvature algorithms suppose that the triangle vertices are oriented clockwise!</b>
 * </p>
 * 
 * @author Natalia Bebjakova
 * @author Radek Oslejsek
 */
public class Curvature extends MeshVisitor {
    
    private final Map<MeshFacet, List<Double>> gaussian = new HashMap<>();
    private final Map<MeshFacet, List<Double>> mean = new HashMap<>();
    private final Map<MeshFacet, List<Double>> minPrincipal = new HashMap<>();
    private final Map<MeshFacet, List<Double>> maxPrincipal = new HashMap<>();
    
    /**
     * Returns Gaussian curvatures for all inspected mesh facets. The order corresponds to
     * the order of vertices, i.e., i-th value represents the curvature of i-th mesh vertex.
     * 
     * @return Gaussian curvatures of inspected mesh facets.
     */
    public Map<MeshFacet, List<Double>> getGaussianCurvatures() {
        return Collections.unmodifiableMap(gaussian);
    }
    
    /**
     * Returns mean curvatures for all inspected mesh facets. The order corresponds to
     * the order of vertices, i.e., i-th value represents the curvature of i-th mesh vertex.
     * 
     * @return Mean curvatures of inspected mesh facets.
     */
    public Map<MeshFacet, List<Double>> getMeanCurvatures() {
        return Collections.unmodifiableMap(mean);
    }
    
    /**
     * Returns minimum principal curvatures for all inspected mesh facets. The order corresponds to
     * the order of vertices, i.e., i-th value represents the curvature of i-th mesh vertex.
     * 
     * @return Minimum principal curvatures of inspected mesh facets.
     */
    public Map<MeshFacet, List<Double>> getMinPrincipalCurvatures() {
        return Collections.unmodifiableMap(minPrincipal);
    }
    
    /**
     * Returns maximum principal curvatures for all inspected mesh facets. The order corresponds to
     * the order of vertices, i.e., i-th value represents the curvature of i-th mesh vertex.
     * 
     * @return Maximum principal curvatures of inspected mesh facets.
     */
    public Map<MeshFacet, List<Double>> getMaxPrincipalCurvatures() {
        return Collections.unmodifiableMap(maxPrincipal);
    }
    
    @Override
    public void visitMeshFacet(final MeshFacet facet) {
        synchronized (this) {
            if (gaussian.containsKey(facet)) {
                return; // already visited facet
            }
            gaussian.put(facet, new ArrayList<>());
            mean.put(facet, new ArrayList<>());
            minPrincipal.put(facet, new ArrayList<>());
            maxPrincipal.put(facet, new ArrayList<>());
        }
        
        Cache cache = new Cache(facet);
        
        final List<CurvTriangle> triangles = precomputeTriangles(facet);
        for (int vertA = 0; vertA < facet.getNumberOfVertices(); vertA++) {
            TriangleFan oneRing = facet.getOneRingNeighborhood(vertA);
            
            if (oneRing.isBoundary()) {
                this.gaussian.get(facet).add(0.0);
                this.mean.get(facet).add(0.0);
                this.minPrincipal.get(facet).add(0.0);
                this.maxPrincipal.get(facet).add(0.0);
                continue;
            }
        
            double sumAngles = 0.0;
            double sumArea = 0.0;
            Vector3d pointSum = new Vector3d();
            
            for (int i = 0; i < oneRing.getVertices().size()-1; i++) {
                int vertB = oneRing.getVertices().get(i);
                int vertC = oneRing.getVertices().get(i+1);
                int tri = oneRing.getTriangles().get(i);
                
                // Angles:
                sumAngles += cache.getAngle(vertA, vertB, vertC, tri);
                
                // Area:
                double alpha = cache.getAngle(vertA, vertB, vertC, tri);
                double beta = cache.getAngle(vertB, vertA, vertC, tri);
                double gamma = cache.getAngle(vertC, vertA, vertB, tri);
                
                Vector3d ab = new Vector3d(facet.getVertex(vertB).getPosition());
                ab.sub(facet.getVertex(vertA).getPosition());
        
                Vector3d ac = new Vector3d(facet.getVertex(vertC).getPosition());
                ac.sub(facet.getVertex(vertA).getPosition());
        
                double lenSqrtAB = ab.lengthSquared();
                double lenSqrtAC = ac.lengthSquared();
                
                double piHalf = Math.PI / 2.0; // 90 degrees
                if (alpha >= piHalf || beta >= piHalf || gamma >= piHalf) { // check for obtuse angle
                    double area = 0.5 * lenSqrtAB * lenSqrtAC * Math.sin(alpha);
                    sumArea += (alpha > piHalf) ? area / 2.0 : area / 4.0;
                } else {
                    double cotGamma = 1.0 / Math.tan(gamma);
                    double cotBeta = 1.0 / Math.tan(beta);
                    sumArea += (lenSqrtAB * cotGamma + lenSqrtAC * cotBeta) / 8.0;
                }
                
                // Laplace:
                int triNext = oneRing.getTriangles().get((i + 1) % oneRing.getTriangles().size());
                int vertD;
                if (i == (oneRing.getVertices().size() - 2)) { // last triangle
                    vertD = oneRing.getVertices().get(1);
                } else {
                    vertD = oneRing.getVertices().get(i + 2);
                }
                ac.scale(cache.getCtan(vertA, vertB, vertC, tri) + cache.getCtan(vertA, vertC, vertD, triNext));
                pointSum.add(ac);
            }
        
            double gaussVal = (2.0 * Math.PI - sumAngles) / sumArea;
            double meanVal = 0.25 * sumArea * pointSum.length();
            double delta = Math.max(0, Math.pow(meanVal, 2) - gaussVal);
            
            this.gaussian.get(facet).add(gaussVal);
            this.mean.get(facet).add(meanVal);
            this.minPrincipal.get(facet).add(meanVal - Math.sqrt(delta));
            this.maxPrincipal.get(facet).add(meanVal + Math.sqrt(delta));
        }
    }
    
    protected List<CurvTriangle> precomputeTriangles(MeshFacet facet) {
        List<CurvTriangle> ret = new ArrayList<>(facet.getNumTriangles());
        for (MeshTriangle tri: facet) {
            ret.add(new CurvTriangle(facet, tri));
        }
        return ret;
    }
    
    /**
     * Helper class that caches triangle characteristics used multiples times during the curvature computation.
     * <ul>
     * <li>A = central point of the 1-ring neighborhood</li>
     * <li>B = point "on the left" (previous point in the clockwise orientation of the triangle</li>
     * <li>C = point "on the right" (next point in the clockwise orientation of the triangle</li>
     * </ul>
     * 
     * @author Radek Oslejsek
     */
    protected class CurvTriangle {
        
        private final MeshFacet facet;
        private final MeshTriangle tri;
        
        private final double v1Angle;
        private final double v2Angle;
        private final double v3Angle;
        
        private final double v1AngleCotang;
        private final double v2AngleCotang;
        private final double v3AngleCotang;
        
        public CurvTriangle(MeshFacet facet, MeshTriangle tri) {
            this.facet = facet;
            this.tri = tri;
            
            Vector3d a = new Vector3d(facet.getVertex(tri.index3).getPosition());
            a.sub(facet.getVertex(tri.index2).getPosition());
            
            Vector3d b = new Vector3d(facet.getVertex(tri.index1).getPosition());
            b.sub(facet.getVertex(tri.index3).getPosition());
            
            Vector3d c = new Vector3d(facet.getVertex(tri.index2).getPosition());
            c.sub(facet.getVertex(tri.index1).getPosition());
            
            a.normalize();
            b.normalize();
            c.normalize();
            
            b.scale(-1.0);
            double cos1 = c.dot(b);
            
            c.scale(-1.0);
            double cos2 = a.dot(c);
            
            a.scale(-1.0);
            b.scale(-1.0);
            double cos3 = a.dot(b);
            
            this.v1Angle = Math.acos(cos1);
            this.v2Angle = Math.acos(cos2);
            this.v3Angle = Math.acos(cos3);
            
            this.v1AngleCotang = 1.0 / Math.tan(v1Angle);
            this.v2AngleCotang = 1.0 / Math.tan(v2Angle);
            this.v3AngleCotang = 1.0 / Math.tan(v3Angle);
        }
        
        public double area() {
            return 0.5 * lengthSquaredAC(tri.vertex1) * lengthSquaredAB(tri.vertex1) * Math.sin(v1Angle);
        }
        
        /**
         * Returns vertex A. 
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return vertex A.
         */
        public Vector3d vertA(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return tri.vertex1.getPosition();
            } else if (facetPoint == tri.vertex2) {
                return tri.vertex2.getPosition();
            } else if (facetPoint == tri.vertex3) {
                return tri.vertex3.getPosition();
            }
            return null;
        }

        /**
         * Returns vertex B. 
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return vertex B.
         */
        public Vector3d vertB(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return tri.vertex3.getPosition();
            } else if (facetPoint == tri.vertex2) {
                return tri.vertex1.getPosition();
            } else if (facetPoint == tri.vertex3) {
                return tri.vertex2.getPosition();
            }
            return null;
        }

        /**
         * Returns vertex C. 
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return vertex C.
         */
        public Vector3d vertC(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return tri.vertex2.getPosition();
            } else if (facetPoint == tri.vertex2) {
                return tri.vertex3.getPosition();
            } else if (facetPoint == tri.vertex3) {
                return tri.vertex1.getPosition();
            }
            return null;
        }

        /**
         * Returns cached angle in the vertex A. 
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return Alpha angle.
         */
        public double alpha(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return v1Angle;
            } else if (facetPoint == tri.vertex2) {
                return v2Angle;
            } else if (facetPoint == tri.vertex3) {
                return v3Angle;
            }
            return Double.NaN;
        }

        /**
         * Returns cached angle in the vertex B.
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return Beta angle.
         */
        public double beta(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return v3Angle;
            } else if (facetPoint == tri.vertex2) {
                return v1Angle;
            } else if (facetPoint == tri.vertex3) {
                return v2Angle;
            }
            return Double.NaN;
        }

        /**
         * Returns cached angle in the vertex C. 
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return Gamma angle.
         */
        public double gamma(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return v2Angle;
            } else if (facetPoint == tri.vertex2) {
                return v3Angle;
            } else if (facetPoint == tri.vertex3) {
                return v1Angle;
            }
            return Double.NaN;
        }

        /**
         * Returns cached cotangent of the angle in the vertex A. 
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return Cotangent of the alpha angle.
         */
        public double alphaCotan(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return v1AngleCotang;
            } else if (facetPoint == tri.vertex2) {
                return v2AngleCotang;
            } else if (facetPoint == tri.vertex3) {
                return v3AngleCotang;
            }
            return Double.NaN;
        }

        /**
         * Returns cached cotangent of the angle in the vertex B.
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return Cotangent of the beta angle.
         */
        public double betaCotan(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return v3AngleCotang;
            } else if (facetPoint == tri.vertex2) {
                return v1AngleCotang;
            } else if (facetPoint == tri.vertex3) {
                return v2AngleCotang;
            }
            return Double.NaN;
        }

        /**
         * Returns cached cotangent of the angle in the vertex C. 
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return Cotangent of the gamma angle.
         */
        public double gammaCotan(MeshPoint facetPoint) {
            if (facetPoint == tri.vertex1) {
                return v2AngleCotang;
            } else if (facetPoint == tri.vertex2) {
                return v3AngleCotang;
            } else if (facetPoint == tri.vertex3) {
                return v1AngleCotang;
            }
            return Double.NaN;
        }
        
        /**
         * Returns squared length of the edge AB (opposite to the vertex A). 
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return Squared length of the edge AB.
         */
        public double lengthSquaredAB(MeshPoint facetPoint) {
            Vector3d v = null;
            if (facetPoint == tri.vertex1) {
                v = new Vector3d(facet.getVertex(tri.index3).getPosition());
                v.sub(facet.getVertex(tri.index1).getPosition());
            } else if (facetPoint == tri.vertex2) {
                v = new Vector3d(facet.getVertex(tri.index1).getPosition());
                v.sub(facet.getVertex(tri.index2).getPosition());
            } else if (facetPoint == tri.vertex3) {
                v = new Vector3d(facet.getVertex(tri.index2).getPosition());
                v.sub(facet.getVertex(tri.index3).getPosition());
            }
            return v.lengthSquared();
        }
        
        /**
         * Returns squared length of the edge AC (opposite to the vertex B). 
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return Squared length of the edge AC.
         */
        public double lengthSquaredAC(MeshPoint facetPoint) {
            Vector3d v = null;
            if (facetPoint == tri.vertex1) {
                v = new Vector3d(facet.getVertex(tri.index2).getPosition());
                v.sub(facet.getVertex(tri.index1).getPosition());
            } else if (facetPoint == tri.vertex2) {
                v = new Vector3d(facet.getVertex(tri.index3).getPosition());
                v.sub(facet.getVertex(tri.index2).getPosition());
            } else if (facetPoint == tri.vertex3) {
                v = new Vector3d(facet.getVertex(tri.index1).getPosition());
                v.sub(facet.getVertex(tri.index3).getPosition());
            }
            return v.lengthSquared();
        }
        
        /**
         * Returns squared length of the edge BC (opposite to the vertex A). 
         * 
         * @param facetPoint Central point of the 1-ring neighborhood
         * @return Squared length of the edge BC.
         */
        public double lengthSquaredBC(MeshPoint facetPoint) {
            Vector3d v = null;
            if (facetPoint == tri.vertex1) {
                v = new Vector3d(facet.getVertex(tri.index3).getPosition());
                v.sub(facet.getVertex(tri.index2).getPosition());
                v.sub(facetPoint.getPosition());
            } else if (facetPoint == tri.vertex2) {
                v = new Vector3d(facet.getVertex(tri.index1).getPosition());
                v.sub(facet.getVertex(tri.index3).getPosition());
            } else if (facetPoint == tri.vertex3) {
                v = new Vector3d(facet.getVertex(tri.index2).getPosition());
                v.sub(facet.getVertex(tri.index1).getPosition());
            }
            return v.lengthSquared();
        }
    }

    /**
     * Cache key.
     * @author Radek Oslejsek
     */
    private class CacheKey {

        private final int vertIndex;
        private final int triIndex;
        
        /**
         * Constructor.
         * @param v vertex
         * @param i triangle
         */
        CacheKey(int v, int i) {
            this.vertIndex = v;
            this.triIndex = i;
        }

        @Override
        public int hashCode() {
            int hash = 7;
            hash = 61 * hash + this.vertIndex;
            hash = 61 * hash + this.triIndex;
            return hash;
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj) {
                return true;
            }
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            final CacheKey other = (CacheKey) obj;
            if (this.vertIndex != other.vertIndex) {
                return false;
            }
            if (this.triIndex != other.triIndex) {
                return false;
            }
            return true;
        }
    }
    
    /**
     * Cache.
     * @author Radek Oslejsek
     */
    private class Cache {
        
        private final MeshFacet facet;
        private final Map<CacheKey, Double> cache = new HashMap<>();

        Cache(MeshFacet facet) {
            this.facet = facet;
        }
        
        public double getAngle(int vertA, int vertB, int vertC, int tri) {
            CacheKey key = new CacheKey(vertA, tri);
            if (cache.containsKey(key)) {
                return cache.get(key);
            }
            
            //System.out.println("XXX"+vertA+"|"+vertB+"|"+vertC);
            //System.out.println("YYY"+facet.getVertex(vertA).getPosition()+"|"+facet.getVertex(vertB).getPosition()+"|"+facet.getVertex(vertC).getPosition());
            
            Vector3d ab = new Vector3d(facet.getVertex(vertB).getPosition());
            ab.sub(facet.getVertex(vertA).getPosition());
            ab.normalize();
            
            Vector3d ac = new Vector3d(facet.getVertex(vertC).getPosition());
            ac.sub(facet.getVertex(vertA).getPosition());
            ac.normalize();
            
            Vector3d bc = new Vector3d(facet.getVertex(vertC).getPosition());
            bc.sub(facet.getVertex(vertB).getPosition());
            bc.normalize();
            
            //System.out.println("AAA"+ac+"|"+ab+"|"+bc);
            
            double cosAlpha = ab.dot(ac);
            
            ab.scale(-1.0);
            double cosBeta = ab.dot(bc);
            
            bc.scale(-1.0);
            ac.scale(-1.0);
            double cosGamma = bc.dot(ac);
            
            //System.out.println("BBB"+cosAlpha+"|"+cosBeta+"|"+cosGamma);
            
            double ret = Math.acos(cosAlpha);
            
            cache.put(new CacheKey(vertA, tri), ret);
            cache.put(new CacheKey(vertB, tri), Math.acos(cosBeta));
            cache.put(new CacheKey(vertC, tri), Math.acos(cosGamma));
            
            return ret;
        }
        
        public double getCtan(int vertA, int vertB, int vertC, int tri) {
            return 1.0 / Math.tan(getAngle(vertA, vertB, vertC, tri));
        }
    }
    

}
